import pytest

from xonsh.completers.tools import RichCompletion
from xonsh.readline_shell import _render_completions


@pytest.mark.parametrize(
    'prefix, completion, prefix_len, readline_completion', [
        ('', 'a', 0, 'a'),
        ('a', 'b', 0, 'ab'),
        ('a', 'b', 1, 'b'),
        ('adc', 'bc', 2, 'abc'),
        ('', RichCompletion('x', 0), 0, 'x'),
        ('', RichCompletion('x', 0, 'aaa', 'aaa'), 0, 'x'),
        ('a', RichCompletion('b', 1), 0, 'b'),
        ('a', RichCompletion('b', 0), 1, 'ab'),
        ('a', RichCompletion('b'), 0, 'ab'),
        ('a', RichCompletion('b'), 1, 'b'),
    ])
def test_render_completions(prefix, completion, prefix_len,
                            readline_completion):
    assert _render_completions({completion}, prefix, prefix_len) == [
        readline_completion]



import pytest
from unittest.mock import MagicMock
from prompt_toolkit.completion import Completion as PTKCompletion

from xonsh.aliases import Aliases
from xonsh.completers.tools import RichCompletion
from xonsh.ptk_shell.completer import PromptToolkitCompleter


@pytest.mark.parametrize('completion, lprefix, ptk_completion', [
    (RichCompletion('x', 0, 'x()', 'func'), 0, None),
    (RichCompletion('x', 1, 'xx', 'instance'), 0, None),
    (RichCompletion('x', description='wow'), 5,
     PTKCompletion(RichCompletion('x'), -5, 'x', 'wow')),
    (RichCompletion('x'), 5, PTKCompletion(RichCompletion('x'), -5, 'x')),
    ('x', 5, PTKCompletion('x', -5, 'x')),
])
def test_rich_completion(completion, lprefix, ptk_completion, monkeypatch):
    xonsh_completer_mock = MagicMock()
    xonsh_completer_mock.complete.return_value = {completion}, lprefix

    ptk_completer = PromptToolkitCompleter(xonsh_completer_mock, None, None)
    ptk_completer.reserve_space = lambda: None
    ptk_completer.suggestion_completion = lambda _, __: None

    document_mock = MagicMock()
    document_mock.text = ''
    document_mock.current_line = ''
    document_mock.cursor_position_col = 0

    monkeypatch.setattr('builtins.aliases', Aliases())

    completions = list(
        ptk_completer.get_completions(document_mock, MagicMock()))
    if isinstance(completion, RichCompletion) and not ptk_completion:
        assert completions == [
            PTKCompletion(completion, -completion.prefix_len,
                          completion.display,
                          completion.description)]
    else:
        assert completions == [ptk_completion]



"""Tests for the Jedi completer xontrib"""
import sys
import pytest
import builtins
import importlib
from unittest.mock import MagicMock, call

from xonsh.xontribs import find_xontrib
from xonsh.completers.tools import RichCompletion


@pytest.fixture
def jedi_mock(monkeypatch):
    jedi_mock = MagicMock()
    jedi_mock.__version__ = '0.16.0'
    jedi_mock.Interpreter().complete.return_value = []
    jedi_mock.reset_mock()
    monkeypatch.setitem(sys.modules, 'jedi', jedi_mock)
    yield jedi_mock


@pytest.fixture
def completer_mock(monkeypatch):
    completer_mock = MagicMock()

    # so that args will be passed
    def comp(args):
        completer_mock(args)

    monkeypatch.setitem(builtins.aliases, 'completer', comp)
    yield completer_mock


@pytest.fixture
def jedi_xontrib(monkeypatch, source_path, jedi_mock, completer_mock):
    monkeypatch.syspath_prepend(source_path)
    spec = find_xontrib('jedi')
    yield importlib.import_module(spec.name)
    del sys.modules[spec.name]


def test_completer_added(jedi_xontrib, completer_mock):
    assert completer_mock.call_args_list == [call(['remove', 'python_mode']),
                                             call(['add', 'jedi_python',
                                                   'complete_jedi', '<python']),
                                             call(['remove', 'python'])]


@pytest.mark.parametrize('prefix, line, start, end, ctx', [
    ('x', '10 + x', 5, 6, {}),
], ids='x')
@pytest.mark.parametrize('version', ['new', 'old'])
def test_jedi_api(jedi_xontrib, jedi_mock, version, prefix, line, start, end,
                  ctx):
    if version == 'old':
        jedi_mock.__version__ = '0.15.0'
        jedi_mock.Interpreter().completions.return_value = []
        jedi_mock.reset_mock()

    jedi_xontrib.complete_jedi(prefix, line, start, end, ctx)

    extra_namespace = {'__xonsh__': builtins.__xonsh__}
    try:
        extra_namespace['_'] = _
    except NameError:
        pass
    namespaces = [{}, extra_namespace]

    if version == 'new':
        assert jedi_mock.Interpreter.call_args_list == [call(line, namespaces)]
        assert jedi_mock.Interpreter().complete.call_args_list == [call(1, end)]
    else:
        assert jedi_mock.Interpreter.call_args_list == [
            call(line, namespaces, line=1, column=end)]
        assert jedi_mock.Interpreter().completions.call_args_list == [call()]


def test_multiline(jedi_xontrib, jedi_mock, monkeypatch):
    shell_mock = MagicMock()
    complete_document = 'xx = 1\n1 + x'
    shell_mock.shell_type = 'prompt_toolkit'
    shell_mock.shell.pt_completer.current_document.text = complete_document
    shell_mock.shell.pt_completer.current_document.cursor_position_row = 1
    shell_mock.shell.pt_completer.current_document.cursor_position_col = 5
    monkeypatch.setattr(builtins.__xonsh__, 'shell', shell_mock)
    jedi_xontrib.complete_jedi('x', 'x', 0, 1, {})

    assert jedi_mock.Interpreter.call_args_list[0][0][0] == complete_document
    assert jedi_mock.Interpreter().complete.call_args_list == [
        call(2, 5)  # line (one-indexed), column (zero-indexed)
    ]


@pytest.mark.parametrize('completion, rich_completion', [
    (
            # from jedi when code is 'x' and xx=3
            ('instance', 'xx', 'x', 'int(x=None, /) -> int',
             ('instance', 'instance int')),
            RichCompletion('x', display='xx', description='instance int')
    ),
    (
            # from jedi when code is 'xx=3\nx'
            ('statement', 'xx', 'x', None, ('instance', 'instance int')),
            RichCompletion('x', display='xx', description='instance int')
    ),
    (
            # from jedi when code is 'x.' and x=3
            ('function', 'from_bytes', 'from_bytes',
             'from_bytes(bytes, byteorder, *, signed=False)',
             ('function', 'def __get__')),
            RichCompletion('from_bytes', display='from_bytes()',
                           description=
                           'from_bytes(bytes, byteorder, *, signed=False)')
    ),
    (
            # from jedi when code is 'x=3\nx.'
            ('function', 'imag', 'imag', None, ('instance', 'instance int')),
            RichCompletion('imag', display='imag', description='instance int')
    ),
    (
            # from '(3).from_bytes(byt'
            ('param', 'bytes=', 'es=', None, ('instance', 'instance Sequence')),
            RichCompletion('es=', display='bytes=',
                           description='instance Sequence')
    ),
    (
            # from 'x.from_bytes(byt' when x=3
            ('param', 'bytes=', 'es=', None, None),
            RichCompletion('es=', display='bytes=', description='param')
    ),
    (
            # from 'import colle'
            ('module', 'collections', 'ctions', None,
             ('module', 'module collections')),
            RichCompletion('ctions', display='collections',
                           description='module collections')
    ),
    (
            # from 'NameErr'
            ('class', 'NameError', 'or', 'NameError(*args: object)',
             ('class', 'class NameError')),
            RichCompletion('or', display='NameError',
                           description='NameError(*args: object)')
    ),
    (
            # from 'a["' when a={'name':None}
            ('string', '"name"', 'name"', None, None),
            RichCompletion('name"', display='"name"', description='string')
    ),
    (
            # from 'open("/etc/pass'
            ('path', 'passwd"', 'wd"', None, None),
            RichCompletion('wd"', display='passwd"', description='path')
    ),
    (
            # from 'cla'
            ('keyword', 'class', 'ss', None, None),
            RichCompletion('ss', display='class', description='keyword')
    ),
])
def test_rich_completions(jedi_xontrib, jedi_mock, completion, rich_completion):
    comp_type, comp_name, comp_complete, sig, inf = completion
    comp_mock = MagicMock()
    comp_mock.type = comp_type
    comp_mock.name = comp_name
    comp_mock.complete = comp_complete
    if sig:
        sig_mock = MagicMock()
        sig_mock.to_string.return_value = sig
        comp_mock.get_signatures.return_value = [sig_mock]
    else:
        comp_mock.get_signatures.return_value = []
    if inf:
        inf_type, inf_desc = inf
        inf_mock = MagicMock()
        inf_mock.type = inf_type
        inf_mock.description = inf_desc
        comp_mock.infer.return_value = [inf_mock]
    else:
        comp_mock.infer.return_value = []

    jedi_xontrib.XONSH_SPECIAL_TOKENS = []
    jedi_mock.Interpreter().complete.return_value = [comp_mock]
    completions = jedi_xontrib.complete_jedi('', '', 0, 0, {})
    assert len(completions) == 1
    ret_completion, = completions
    assert isinstance(ret_completion, RichCompletion)
    assert ret_completion == rich_completion
    assert ret_completion.display == rich_completion.display
    assert ret_completion.description == rich_completion.description


def test_special_tokens(jedi_xontrib):
    assert jedi_xontrib.complete_jedi('', '', 0, 0,
                                      {}) == jedi_xontrib.XONSH_SPECIAL_TOKENS
    assert jedi_xontrib.complete_jedi('@', '@', 0, 1, {}) == {'@', '@(', '@$('}
    assert jedi_xontrib.complete_jedi('$', '$', 0, 1, {}) == {'$[', '${', '$('}



# -*- coding: utf-8 -*-
"""Test initialization of prompt_toolkit shell"""

import pytest

from xonsh.platform import minimum_required_ptk_version

# verify error if ptk not installed or below min

from xonsh.shell import Shell


@pytest.mark.parametrize(
    "ptk_ver, ini_shell_type, exp_shell_type, warn_snip",
    [
        (None, "prompt_toolkit", "readline", "prompt_toolkit is not available"),
        ((0, 5, 7), "prompt_toolkit", "readline", "is not supported"),
        ((1, 0, 0), "prompt_toolkit", "readline", "is not supported"),
        ((2, 0, 0), "prompt_toolkit", "prompt_toolkit", None),
        ((2, 0, 0), "best", "prompt_toolkit", None),
        ((2, 0, 0), "readline", "readline", None),
        ((3, 0, 0), "prompt_toolkit", "prompt_toolkit", None),
        ((3, 0, 0), "best", "prompt_toolkit", None),
        ((3, 0, 0), "readline", "readline", None),
        ((4, 0, 0), "prompt_toolkit", "prompt_toolkit", None),
    ],
)
def test_prompt_toolkit_version_checks(ptk_ver, ini_shell_type, exp_shell_type, warn_snip, monkeypatch, xonsh_builtins):

    mocked_warn = ""

    def mock_warning(msg):
        nonlocal mocked_warn
        mocked_warn = msg
        return

    def mock_ptk_above_min_supported():
        nonlocal ptk_ver
        return ptk_ver and (ptk_ver[:2] >= minimum_required_ptk_version)

    def mock_has_prompt_toolkit():
        nonlocal ptk_ver
        return ptk_ver is not None

    monkeypatch.setattr("xonsh.shell.warnings.warn", mock_warning)      # hardwon: patch the caller!
    monkeypatch.setattr("xonsh.shell.ptk_above_min_supported", mock_ptk_above_min_supported)    # have to patch both callers
    monkeypatch.setattr("xonsh.platform.ptk_above_min_supported", mock_ptk_above_min_supported)
    monkeypatch.setattr("xonsh.shell.has_prompt_toolkit", mock_has_prompt_toolkit)
    monkeypatch.setattr("xonsh.platform.has_prompt_toolkit", mock_has_prompt_toolkit)

    act_shell_type = Shell.choose_shell_type(ini_shell_type, {})

    assert act_shell_type == exp_shell_type

    if warn_snip:
        assert warn_snip in mocked_warn

    pass

# someday: initialize PromptToolkitShell and have it actually do something.
